(* Version 2.1 *)
program HpglInterpreter;
(*$IKERNEL.INC*)
(*$IGRAPHLIB.INC*)
label
  byebye;
const
  TEST = false;
  PAUSE_MODE = true;
  START     = 'S';
  MNEMONIC  = 'M';
  PARAMETER = 'P';
  ERROR     = 'E';
  UNKNOWN   = 0;
  INIT      = 1;
  PEN_DOWN  = 2;
  PEN_UP    = 3;
  POS_ABS   = 4;
  POS_REL   = 5;
  MIN_COORD = -10000;
  MAX_COORD =  10000;
var
  Break:       Boolean;
  FilVar:      Text;
  Ch:          Char;
  State:       Char;
  I:           Integer;
  FileName:    string(.14.);
  Command:     Integer;
  FirstMnemCh: Char;
  Value:       Integer;
  IsNegative:  Boolean;
  IsPenDown:   Boolean;
  LastX:       Integer;
  LastY:       Integer;
  ParamX:      Integer;
  ParamY:      Integer;
  ParamName:   Char;
  DotCnt:      Integer;
  MaxX, MaxY:  Integer;
  MinX, MinY:  Integer;
  OffsetX:     Integer;
  OffsetY:     Integer;
  ScaleX:      Real;
  ScaleY:      Real;
  Pass:        Integer;


(* forward declarations *)
procedure PrintCommand; forward;
procedure Draw;         forward;


(* functions and procedures *)
procedure InitParameter;
begin
  Value := 0;
  IsNegative := false;
end; (* InitParameter *)


procedure Beep;
begin
  Write(chr(7));
end; (* Beep *)


function isWhitespace(c: Char): Boolean;
begin
  isWhitespace := ord(c) < 33;
end; (* isWhitespace *)


procedure FinishParameter;
begin
  if (Command > INIT) and (Command < 6) then begin
    if IsNegative then Value := -Value;
    if ParamName = 'X' then begin
      LastX := ParamX;
      ParamX := Value;
      if Value > MaxX then MaxX := Value;
      if Value < MinX then MinX := Value;
      ParamName := 'Y';
    end else begin
      LastY := ParamY;
      ParamY := Value;
      if Value > MaxY then MaxY := Value;
      if Value < MinY then MinY := Value;
      ParamName := 'X';
      if (Pass = 2) and IsPenDown then Draw;
    end;
  end;
  InitParameter;
end; (* FinishParameter *)


procedure ScaleParameter;
var r: Real;
begin
  r := ParamX - MinX;
  r := r * ScaleX;
  ParamX := trunc(r) + OffsetX;

  r := MaxY - ParamY; (* + MinY ??? *)
  r := r * ScaleY;
  ParamY := trunc(r) + OffsetY;
end; (* ScaleParameter *)


procedure PrintDot;
begin
  if Pass = 1 then begin
    DotCnt := DotCnt - 1;
    if DotCnt = 0 then begin
      DotCnt := 10;
      Write('.');
    end;
  end;
end; (* PrintDot *)


procedure ProcessChar;
begin
  if Ch = ';' then begin
    State := START;
    FinishParameter;
    ParamName := 'X';
  end else begin
    case State of
      START: begin
          PrintDot;
          Command := UNKNOWN;
          InitParameter;
          if not isWhitespace(Ch) then begin
            if (Ch >= 'A') and (Ch <= 'Z') then begin
              FirstMnemCh := Ch;
              State := MNEMONIC;
            end else begin
              Writeln;
              Writeln('Invalid first character in Mnemonic: ', Ch);
              State := ERROR;
            end;
          end;
        end;
      MNEMONIC: begin
          if (Ch >= 'A') and (Ch <= 'Z') then begin
            case FirstMnemCh of
              'I': case Ch of
                     'N': begin Command := INIT;     IsPenDown := false; end;
                   end;
              'P': case Ch of
                     'D': begin Command := PEN_DOWN; IsPenDown := true;  end;
                     'U': begin Command := PEN_UP;   IsPenDown := false; end;
                     'A': begin Command := POS_ABS;                      end;
                     'R': begin Command := POS_REL;                      end;
                   end;
            end;
            State := PARAMETER;
          end else begin
            Writeln;
            Writeln('Invalid second character in Mnemonic: ', Ch);
            State := ERROR;
          end;
        end;
      PARAMETER: begin
          if Ch <> '.' then begin (* real values not supported; dot ignored *)
            if (Ch = '-') then begin
              IsNegative := true;
            end else if (Ch >= '0') and (Ch <='9') then begin
              Value := Value * 10 + Ord(Ch) - Ord('0');
            end else begin
              (* delimiter found: end parameter *)
              FinishParameter;
            end;
          end;
        end;
    end; (* case *)
  end;
end; (* ProcessChar *)


procedure PrintCommand;
begin
  case Command of
    UNKNOWN:  Write('UNKNOWN');
    PEN_DOWN: Write('PEN_DOWN');
    PEN_UP:   Write('PEN_UP');
    POS_ABS:  Write('POS_ABS');
    POS_REL:  Write('POS_REL');
  end;
end; (* PrintCommand *)


procedure Draw;
begin
  ScaleParameter;
  if (LastX >= 0) and (LastX < 720) and
     (LastY >= 0) and (LastY < 256) and
     (ParamX >= 0) and (ParamX < 720) and
     (ParamY >= 0) and (ParamY < 256) then begin
    line(LastX, LastY, ParamX, ParamY, 0);
  end;
end; (* Draw *)


procedure CalcScaling;
var height, width: Integer;
    h2, w2: Real;
begin
  width  := MaxX - MinX;
  height := MaxY - MinY;
  if height > width then begin
    ScaleX := 540 / height;
    ScaleY := ScaleX * 256 / 540;
    w2 := width * ScaleX;
    OffsetX := (720 - trunc(w2)) div 2;
    OffsetY := 0;
  end else begin
    ScaleY := 720 / width;
    ScaleX := ScaleY * 540 / 256;
    h2 := height * ScaleY;
    OffsetX := 0;
    OffsetY := (256 - trunc(h2)) div 2;
  end;
end; (* CalcScaling *)


(* Main program *)
begin
  ClrScr;
  if TEST then begin
    FileName := 'g2.plt';
  end
  else begin
    Write('Enter name of file to draw: ');
    Readln(FileName);
  end;
  Assign(FilVar, FileName);
  (*$I-*)
  Reset(FilVar);
  (*$I+*)
  if IOResult <> 0 then
  begin
    Writeln('Datei nicht gefunden: ', FileName);
  end;

  Break := false;
  State := START;
  ParamName := 'X';
  LastX := 0;
  LastY := 0;
  MinX := MAX_COORD;
  MaxX := MIN_COORD;
  MinY := MAX_COORD;
  MaxY := MIN_COORD;
  DotCnt := 10;

  ClrScr;
  HideCursor;
  Writeln('First pass: determining min. and max. X and Y coordinates.');

  for Pass := 1 to 2 do begin
    if Pass = 2 then begin
      ClrScr;
      Writeln('Second pass: drawing');
      Writeln;
      Writeln('Min. X: ', MinX, '   Max. X: ', MaxX);
      Writeln('Min. Y: ', MinY, '   Max. Y: ', MaxY);
      Writeln;
      CalcScaling;
      Writeln('Scale  X: ', ScaleX,  '   Scale  Y: ', ScaleY);
      Writeln('Offset X: ', OffsetX, '   Offset Y: ', OffsetY);
      Repeat until KeyPressed;

      Reset(FilVar);
      ClrScr;
      GraphInit;
    end;

    while (not Eof(FilVar)) and (State <> ERROR) do begin
      Read(FilVar, Ch);
      ProcessChar;
    end; (* while *)
  end; (* for *)

  Beep;
  repeat until KeyPressed;

byebye:
  ShowCursor;
end.
